@@ -43,7 +43,7 @@ WindowAttentionHandler.prototype = {
             }
         }
         catch (e) {
-            global.logError('' + e);
+            global.logError('Error showing window demanding attention', e);
         }
     }
 };
